import pytest

from prefect_firebolt import FireboltCredentials


@pytest.mark.parametrize(
    "username,password,token,error_message",
    [
        ("me@example.com", "hunter2", None, None),
        (
            "me@example.com",
            None,
            None,
            "You have provided only a username or password.",
        ),
        (None, "hunter2", None, "You have provided only a username or password."),
        (
            None,
            "hunter2",
            "abc123",
            "You have provided both a username/password and a token.",
        ),
        (None, None, "abc123", None),
        (None, None, None, "You have not provided a username/password or token."),
    ],
)
def test_validate_auth_parameter_combinations(username, password, token, error_message):
    if error_message:
        with pytest.raises(ValueError, match=error_message):
            FireboltCredentials(username=username, password=password, token=token)
    else:
        FireboltCredentials(username=username, password=password, token=token)



"""Module for interacting with Firebolt databases"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence

from firebolt.async_db.connection import Connection
from prefect import task
from prefect.blocks.core import Block
from prefect.utilities.asyncutils import sync_compatible
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import Field, root_validator
else:
    from pydantic import Field, root_validator

from prefect_firebolt.credentials import FireboltCredentials

if TYPE_CHECKING:
    from firebolt.async_db._types import ColType, ParameterType


class FireboltDatabase(Block):
    """
    Connects to a Firebolt database.

    Provide either `engine_name` or `engine_url`. Providing both will result in an
    error. If neither `engine_name` nor `engine_url` is provided, the default engine
    for the configured database will be used.

    Attributes:
        credentials: Credentials to use to connect to the Firebase database.
        database_name: The name of the database to connect to.
        engine_name: Name of the engine to connect to.
        engine_url: The engine endpoint to use.
        additional_parameters: Additional configuration to pass to the Firebolt
            connection.
    """

    _block_type_name = "Firebolt Database"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/3loU17IXqVIWl4aWQfqc78/3c7eefe5e8cf4eec870856f10d7fdcce/5e8a264ceaf4870c90478037_Favicon_128.svg.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-firebolt/database/#prefect_firebolt.database.FireboltDatabase"  # noqa
    _description = "Connects to a Firebolt database."

    credentials: FireboltCredentials
    database: str = Field(
        default=...,
        title="Database Name",
        description="The name of the database to connect to.",
    )
    engine_name: Optional[str] = Field(
        default=None,
        description="Name of the engine to connect to. May not be used with "
        "engine_url. If neither engine_name nor engine_url is provided, the "
        "default engine for the configured database will be used.",
    )
    engine_url: Optional[str] = Field(
        default=None,
        title="Engine URL",
        description="The engine endpoint to use. May not be used with engine_name. "
        "If neither engine_name nor engine_url is provided, the "
        "default engine for the configured database will be used.",
    )
    additional_parameters: Optional[Dict] = Field(
        default_factory=dict,
        description="Additional configuration to pass to the Firebolt connection.",
    )

    @root_validator(pre=True)
    def _not_both_engine_name_and_engine_url(cls, values):
        """Ensures that engine_name and engine_url are not both provided"""
        if (
            values.get("engine_name") is not None
            and values.get("engine_url") is not None
        ):
            raise ValueError(
                "You have provided a value for both engine_name and engine_url. "
                "Please provide either engine_name or engine_url, but not both."
            )
        return values

    def _get_connect_params(self) -> Dict[str, Any]:
        """Assemble the parameters to pass to the Firebolt connection"""
        return {
            "database": self.database,
            "engine_name": self.engine_name,
            "engine_url": self.engine_url,
            "additional_parameters": self.additional_parameters,
        }

    @sync_compatible
    async def get_connection(self) -> Connection:
        """
        Creates and returns an authenticated Firebolt connection for the
        configured database.
        """
        return await self.credentials.get_client(**self._get_connect_params())


@task
async def query_firebolt(
    database: FireboltDatabase,
    query: str,
    parameters: "Optional[Sequence[ParameterType]]" = None,
) -> "List[List[ColType]]":
    """
    Executes a query against a Firebolt database.

    Args:
        database: Firebolt database configuration to use for query execution.
        query: SQL query to execute.
        parameters: A sequence of substitution parameters. Used to replace `?`
            placeholders inside a query with actual values.

    Returns:
        All rows retrieved by the query.

    Example:
        Execute a query against a Firebolt database:
        ```python
        from prefect import flow

        from prefect_firebolt import FireboltCredentials, FireboltDatabase, query_firebolt


        @flow
        def run_firebolt_query():
            firebolt_database_block = FireboltDatabase(
                database="travel",
                credentials=FireboltCredentials(
                    username="arthur.dent@hitchhikers.com", password="dont42panic"
                ),
            )

            results = query_firebolt(
                database=firebolt_database_block,
                query="SELECT * FROM ex_intergalactic_trips LIMIT 100",
            )

            return results


        run_firebolt_query()
        ```
    """  # noqa
    async with await database.get_connection() as connection:
        cursor = connection.cursor()
        with cursor:
            await cursor.execute(query=query, parameters=parameters)
            return await cursor.fetchall()



"""Module for managing Firebolt credentials."""

from typing import Dict, Optional

from firebolt.async_db.connection import Connection, connect
from firebolt.client.auth import Token, UsernamePassword
from firebolt.client.constants import DEFAULT_API_URL
from prefect.blocks.abstract import CredentialsBlock
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import Field, SecretStr, root_validator
else:
    from pydantic import Field, SecretStr, root_validator


class FireboltCredentials(CredentialsBlock):
    """
    Store credentials for authenticating with Firebolt.

    Attributes:
        username: The email address associated with your Firebolt user.
        password: The password used for connecting to Firebolt.
        token: Authentication token to use instead of username and password.
        api_endpoint: Firebolt API endpoint used for authentication.
        account_name: Name of the account to authenticate with.
            If not provided, the default account will be used.
    """

    _block_type_name = "Firebolt Credentials"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/3loU17IXqVIWl4aWQfqc78/3c7eefe5e8cf4eec870856f10d7fdcce/5e8a264ceaf4870c90478037_Favicon_128.svg.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-firebolt/credentials/#prefect_firebolt.credentials.FireboltCredentials"  # noqa

    @root_validator
    def _ensure_valid_auth_method(cls, values):
        """
        Checks to make sure that either username and password or a token is provided.
        """
        has_username = values.get("username") is not None
        has_password = values.get("password") is not None
        has_token = values.get("token") is not None

        # No auth details provided
        if not any([has_password, has_username, has_token]):
            raise ValueError(
                "You have not provided a username/password or token. "
                "Please provide either a username and password or a token."
            )
        # Username/password and token provided
        if (has_username or has_password) and has_token:
            raise ValueError(
                "You have provided both a username/password and a token. "
                "Please provide either a username and password or a token."
            )
        # Username/password not provided together
        if has_username ^ has_password:
            raise ValueError(
                "You have provided only a username or password. "
                "Please provide both a username and a password."
            )
        return values

    username: Optional[str] = Field(
        default=None,
        description="The email address associated with your Firebolt user.",
    )
    password: Optional[SecretStr] = Field(
        default=None, description="The password used for connecting to Firebolt."
    )
    token: Optional[SecretStr] = Field(
        default=None,
        description="Authentication token to use instead of username and password.",
    )
    api_endpoint: str = Field(
        default=DEFAULT_API_URL,
        title="API Endpoint",
        description="Firebolt API endpoint used for authentication.",
    )
    account_name: Optional[str] = Field(
        default=None,
        description="Name of the account to authenticate with. If not "
        "provided, the default account will be used.",
    )

    async def get_client(self, **connect_params: Dict) -> Connection:
        """Retrieves an initialized Firebolt connection object.

        Args:
            **connect_params: Additional parameters to pass to the Firebolt
                connection object.

        Returns:
            A Firebolt connection object.
        """
        if self.token:
            auth = Token(token=self.token.get_secret_value())
        elif self.username and self.password:
            auth = UsernamePassword(
                username=self.username,
                password=self.password.get_secret_value(),
            )
        else:
            raise ValueError(
                "Unable to initialize Firebolt auth. Expected username "
                "and password or token, but received neither."
            )

        return await connect(
            api_endpoint=self.api_endpoint, auth=auth, **connect_params
        )



import pytest
from prefect.testing.standard_test_suites import BlockStandardTestSuite

from prefect_firebolt import FireboltCredentials
from prefect_firebolt.database import FireboltDatabase


@pytest.mark.parametrize("block", [FireboltCredentials, FireboltDatabase])
class TestAllBlocksAdhereToStandards(BlockStandardTestSuite):
    @pytest.fixture
    def block(self, block):
        return block



from . import _version
from .credentials import FireboltCredentials
from .database import FireboltDatabase, query_firebolt

__version__ = _version.get_versions()["version"]
__all__ = ["FireboltCredentials", "FireboltDatabase", "query_firebolt"]
